@@ -1,53 +1,56 @@
-/*
- * Licensed to the Apache Software Foundation (ASF) under one or more
- * contributor license agreements.  See the NOTICE file distributed with
- * this work for additional information regarding copyright ownership.
- * The ASF licenses this file to You under the Apache License, Version 2.0
- * (the "License"); you may not use this file except in compliance with
- * the License.  You may obtain a copy of the License at
- *
- *     http://www.apache.org/licenses/LICENSE-2.0
- *
- * Unless required by applicable law or agreed to in writing, software
- * distributed under the License is distributed on an "AS IS" BASIS,
- * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
- * See the License for the specific language governing permissions and
- * limitations under the License.
- */
+/*******************************************************************************
+ * Copyright (c) 2015 Voyager Search and MITRE
+ * All rights reserved. This program and the accompanying materials
+ * are made available under the terms of the Apache License, Version 2.0 which
+ * accompanies this distribution and is available at
+ *    http://www.apache.org/licenses/LICENSE-2.0.txt
+ ******************************************************************************/
 
 package com.spatial4j.core.shape.jts;
 
 import com.spatial4j.core.context.SpatialContext;
 import com.spatial4j.core.context.jts.JtsSpatialContext;
-import com.spatial4j.core.distance.DistanceUtils;
 import com.spatial4j.core.exception.InvalidShapeException;
 import com.spatial4j.core.shape.*;
 import com.spatial4j.core.shape.Point;
+import com.spatial4j.core.shape.impl.BBoxCalculator;
+import com.spatial4j.core.shape.impl.BufferedLineString;
 import com.spatial4j.core.shape.impl.PointImpl;
 import com.spatial4j.core.shape.impl.RectangleImpl;
 import com.vividsolutions.jts.geom.*;
+import com.vividsolutions.jts.geom.prep.PreparedGeometry;
+import com.vividsolutions.jts.geom.prep.PreparedGeometryFactory;
 import com.vividsolutions.jts.operation.union.UnaryUnionOp;
 import com.vividsolutions.jts.operation.valid.IsValidOp;
 
 import java.util.ArrayList;
+import java.util.Collection;
 import java.util.List;
 
 /**
  * Wraps a JTS {@link Geometry} (i.e. may be a polygon or basically anything).
- * JTS's does a great deal of the hard work, but there is work here in handling
+ * JTS does a great deal of the hard work, but there is work here in handling
  * dateline wrap.
  */
-public class JtsGeometry implements Shape {
+public class JtsGeometry extends BaseShape<JtsSpatialContext> {
+  /** System property boolean that can disable auto validation in an assert. */
+  public static final String SYSPROP_ASSERT_VALIDATE = "spatial4j.JtsGeometry.assertValidate";
+
   private final Geometry geom;//cannot be a direct instance of GeometryCollection as it doesn't support relate()
   private final boolean hasArea;
   private final Rectangle bbox;
-  private final JtsSpatialContext ctx;
+  protected PreparedGeometry preparedGeometry;
+  protected boolean validated = false;
 
-  public JtsGeometry(Geometry geom, JtsSpatialContext ctx, boolean dateline180Check) {
-    this.ctx = ctx;
+  public JtsGeometry(Geometry geom, JtsSpatialContext ctx, boolean dateline180Check, boolean allowMultiOverlap) {
+    super(ctx);
     //GeometryCollection isn't supported in relate()
-    if (geom.getClass().equals(GeometryCollection.class))
-      throw new IllegalArgumentException("JtsGeometry does not support GeometryCollection but does support its subclasses.");
+    if (geom.getClass().equals(GeometryCollection.class)) {
+      geom = narrowCollectionIfPossible((GeometryCollection)geom);
+      if (geom == null) {
+        throw new IllegalArgumentException("JtsGeometry does not support GeometryCollection but does support its subclasses.");
+      }
+    }
 
     //NOTE: All this logic is fairly expensive. There are some short-circuit checks though.
     if (ctx.isGeo()) {
@@ -55,56 +58,120 @@ public class JtsGeometry implements Shape {
       if (dateline180Check)
         unwrapDateline(geom);//potentially modifies geom
       //If given multiple overlapping polygons, fix it by union
-      geom = unionGeometryCollection(geom);//returns same or new geom
-      Envelope unwrappedEnv = geom.getEnvelopeInternal();
+      if (allowMultiOverlap)
+        geom = unionGeometryCollection(geom);//returns same or new geom
 
       //Cuts an unwrapped geometry back into overlaid pages in the standard geo bounds.
       geom = cutUnwrappedGeomInto360(geom);//returns same or new geom
       assert geom.getEnvelopeInternal().getWidth() <= 360;
       assert ! geom.getClass().equals(GeometryCollection.class) : "GeometryCollection unsupported";//double check
 
-      //note: this bbox may be sub-optimal. If geom is a collection of things near the dateline on both sides then
-      // the bbox will needlessly span most or all of the globe longitudinally.
-      // TODO so consider using MultiShape's planned minimal geo bounding box algorithm once implemented.
-      double envWidth = unwrappedEnv.getWidth();
-      //adjust minX and maxX considering the dateline and world wrap
-      double minX, maxX;
-      if (envWidth >= 360) {
-        minX = -180;
-        maxX = 180;
-      } else {
-        minX = unwrappedEnv.getMinX();
-        maxX = DistanceUtils.normLonDEG(unwrappedEnv.getMinX() + envWidth);
-      }
-      bbox = new RectangleImpl(minX, maxX, unwrappedEnv.getMinY(), unwrappedEnv.getMaxY(), ctx);
+      //Compute bbox
+      bbox = computeGeoBBox(geom);
     } else {//not geo
+      //If given multiple overlapping polygons, fix it by union
+      if (allowMultiOverlap)
+        geom = unionGeometryCollection(geom);//returns same or new geom
+
       Envelope env = geom.getEnvelopeInternal();
-      bbox = new RectangleImpl(env.getMinX(),env.getMaxX(),env.getMinY(),env.getMaxY(), ctx);
+      bbox = new RectangleImpl(env.getMinX(), env.getMaxX(), env.getMinY(), env.getMaxY(), ctx);
     }
     geom.getEnvelopeInternal();//ensure envelope is cached internally, which is lazy evaluated. Keeps this thread-safe.
 
-    //Check geom validity; use helpful error
-    // TODO add way to conditionally skip at your peril later
-    IsValidOp isValidOp = new IsValidOp(geom);
-    if (!isValidOp.isValid())
-      throw new InvalidShapeException(isValidOp.getValidationError().toString());
     this.geom = geom;
+    assert assertValidate();//kinda expensive but caches valid state
 
     this.hasArea = !((geom instanceof Lineal) || (geom instanceof Puntal));
   }
 
-  public static SpatialRelation intersectionMatrixToSpatialRelation(IntersectionMatrix matrix) {
-    if (matrix.isContains())
-      return SpatialRelation.CONTAINS;
-    else if (matrix.isCoveredBy())
-      return SpatialRelation.WITHIN;
-    else if (matrix.isDisjoint())
-      return SpatialRelation.DISJOINT;
-    return SpatialRelation.INTERSECTS;
+  /**
+   * Attempts to retype a geometry collection under the following circumstances, returning
+   * null if the collection can not be retyped.
+   * <ul>
+   *    <li>Single object collections are collapsed down to the object.</li>
+   *    <li>Homogenous collections are recast as the appropriate subclass.</li>
+   * </ul>
+   *
+   * @see GeometryFactory#buildGeometry(Collection)
+   */
+  private Geometry narrowCollectionIfPossible(GeometryCollection gc) {
+    List<Geometry> geoms = new ArrayList<>();
+    for (int i = 0; i < gc.getNumGeometries(); i++) {
+      geoms.add(gc.getGeometryN(i));
+    }
+
+    Geometry result = gc.getFactory().buildGeometry(geoms);
+    return !result.getClass().equals(GeometryCollection.class) ? result : null;
   }
 
-  //----------------------------------------
-  //----------------------------------------
+  /** called via assertion */
+  private boolean assertValidate() {
+    String assertValidate = System.getProperty(SYSPROP_ASSERT_VALIDATE);
+    if (assertValidate == null || Boolean.parseBoolean(assertValidate))
+      validate();
+    return true;
+  }
+
+  /**
+   * Validates the shape, throwing a descriptive error if it isn't valid. Note that this
+   * is usually called automatically by default, but that can be disabled.
+   *
+   * @throws InvalidShapeException with descriptive error if the shape isn't valid
+   */
+  public void validate() throws InvalidShapeException {
+    if (!validated) {
+      IsValidOp isValidOp = new IsValidOp(geom);
+      if (!isValidOp.isValid())
+        throw new InvalidShapeException(isValidOp.getValidationError().toString());
+      validated = true;
+    }
+  }
+
+  /**
+   * Adds an index to this class internally to compute spatial relations faster. In JTS this
+   * is called a {@link com.vividsolutions.jts.geom.prep.PreparedGeometry}.  This
+   * isn't done by default because it takes some time to do the optimization, and it uses more
+   * memory.  Calling this method isn't thread-safe so be careful when this is done. If it was
+   * already indexed then nothing happens.
+   */
+  public void index() {
+    if (preparedGeometry == null)
+      preparedGeometry = PreparedGeometryFactory.prepare(geom);
+  }
+
+  @Override
+  public boolean isEmpty() {
+    return geom.isEmpty();
+  }
+
+  /** Given {@code geoms} which has already been checked for being in world
+   * bounds, return the minimal longitude range of the bounding box.
+   */
+  protected Rectangle computeGeoBBox(Geometry geoms) {
+    if (geoms.isEmpty())
+      return new RectangleImpl(Double.NaN, Double.NaN, Double.NaN, Double.NaN, ctx);
+    final Envelope env = geoms.getEnvelopeInternal();//for minY & maxY (simple)
+    if (ctx.isGeo() && env.getWidth() > 180 && geoms.getNumGeometries() > 1)  {
+      // This is ShapeCollection's bbox algorithm
+      BBoxCalculator bboxCalc = new BBoxCalculator(ctx);
+      for (int i = 0; i < geoms.getNumGeometries(); i++ ) {
+        Envelope envI = geoms.getGeometryN(i).getEnvelopeInternal();
+        bboxCalc.expandXRange(envI.getMinX(), envI.getMaxX());
+        if (bboxCalc.doesXWorldWrap())
+          break; // can't grow any bigger
+      }
+      return new RectangleImpl(bboxCalc.getMinX(), bboxCalc.getMaxX(), env.getMinY(), env.getMaxY(), ctx);
+    } else {
+      return new RectangleImpl(env.getMinX(), env.getMaxX(), env.getMinY(), env.getMaxY(), ctx);
+    }
+  }
+
+  @Override
+  public JtsGeometry getBuffered(double distance, SpatialContext ctx) {
+    //TODO doesn't work correctly across the dateline. The buffering needs to happen
+    // when it's transiently unrolled, prior to being sliced.
+    return this.ctx.makeShape(geom.buffer(distance), true, true);
+  }
 
   @Override
   public boolean hasArea() {
@@ -132,6 +199,8 @@ public class JtsGeometry implements Shape {
 
   @Override
   public JtsPoint getCenter() {
+    if (isEmpty()) //geom.getCentroid == null
+      return new JtsPoint(ctx.getGeometryFactory().createPoint((Coordinate)null), ctx);
     return new JtsPoint(geom.getCentroid(), ctx);
   }
 
@@ -145,21 +214,28 @@ public class JtsGeometry implements Shape {
       return relate((Circle) other);
     else if (other instanceof JtsGeometry)
       return relate((JtsGeometry) other);
+    else if (other instanceof BufferedLineString)
+      throw new UnsupportedOperationException("Can't use BufferedLineString with JtsGeometry");
     return other.relate(this).transpose();
   }
 
   public SpatialRelation relate(Point pt) {
-    //TODO if not jtsPoint, test against bbox to avoid JTS if disjoint
-    JtsPoint jtsPoint = (JtsPoint) (pt instanceof JtsPoint ? pt : ctx.makePoint(pt.getX(), pt.getY()));
-    return geom.disjoint(jtsPoint.getGeom()) ? SpatialRelation.DISJOINT : SpatialRelation.CONTAINS;
+    if (!getBoundingBox().relate(pt).intersects())
+      return SpatialRelation.DISJOINT;
+    Geometry ptGeom;
+    if (pt instanceof JtsPoint)
+      ptGeom = ((JtsPoint)pt).getGeom();
+    else
+      ptGeom = ctx.getGeometryFactory().createPoint(new Coordinate(pt.getX(), pt.getY()));
+    return relate(ptGeom);//is point-optimized
   }
 
   public SpatialRelation relate(Rectangle rectangle) {
     SpatialRelation bboxR = bbox.relate(rectangle);
     if (bboxR == SpatialRelation.WITHIN || bboxR == SpatialRelation.DISJOINT)
       return bboxR;
-    Geometry oGeom = ctx.getGeometryFrom(rectangle);
-    return intersectionMatrixToSpatialRelation(geom.relate(oGeom));
+    // FYI, the right answer could still be DISJOINT or WITHIN, but we don't know yet.
+    return relate(ctx.getGeometryFrom(rectangle));
   }
 
   public SpatialRelation relate(Circle circle) {
@@ -189,9 +265,38 @@ public class JtsGeometry implements Shape {
   }
 
   public SpatialRelation relate(JtsGeometry jtsGeometry) {
-    Geometry oGeom = jtsGeometry.geom;
     //don't bother checking bbox since geom.relate() does this already
-    return intersectionMatrixToSpatialRelation(geom.relate(oGeom));
+    return relate(jtsGeometry.geom);
+  }
+
+  protected SpatialRelation relate(Geometry oGeom) {
+    //see http://docs.geotools.org/latest/userguide/library/jts/dim9.html#preparedgeometry
+    if (oGeom instanceof com.vividsolutions.jts.geom.Point) {
+      if (preparedGeometry != null)
+        return preparedGeometry.disjoint(oGeom) ? SpatialRelation.DISJOINT : SpatialRelation.CONTAINS;
+      return geom.disjoint(oGeom) ? SpatialRelation.DISJOINT : SpatialRelation.CONTAINS;
+    }
+    if (preparedGeometry == null)
+      return intersectionMatrixToSpatialRelation(geom.relate(oGeom));
+    else if (preparedGeometry.covers(oGeom))
+      return SpatialRelation.CONTAINS;
+    else if (preparedGeometry.coveredBy(oGeom))
+      return SpatialRelation.WITHIN;
+    else if (preparedGeometry.intersects(oGeom))
+      return SpatialRelation.INTERSECTS;
+    return SpatialRelation.DISJOINT;
+  }
+
+  public static SpatialRelation intersectionMatrixToSpatialRelation(IntersectionMatrix matrix) {
+    //As indicated in SpatialRelation javadocs, Spatial4j CONTAINS & WITHIN are
+    // OGC's COVERS & COVEREDBY
+    if (matrix.isCovers())
+      return SpatialRelation.CONTAINS;
+    else if (matrix.isCoveredBy())
+      return SpatialRelation.WITHIN;
+    else if (matrix.isDisjoint())
+      return SpatialRelation.DISJOINT;
+    return SpatialRelation.INTERSECTS;
   }
 
   @Override
@@ -229,7 +334,7 @@ public class JtsGeometry implements Shape {
   private static int unwrapDateline(Geometry geom) {
     if (geom.getEnvelopeInternal().getWidth() < 180)
       return 0;//can't possibly cross the dateline
-    final int[] result = {0};//an array so that an inner class can modify it.
+    final int[] crossings = {0};//an array so that an inner class can modify it.
     geom.apply(new GeometryFilter() {
       @Override
       public void filter(Geometry geom) {
@@ -244,12 +349,13 @@ public class JtsGeometry implements Shape {
           cross = unwrapDateline((Polygon) geom);
         } else
           return;
-        result[0] = Math.max(result[0], cross);
+        crossings[0] = Math.max(crossings[0], cross);
       }
     });//geom.apply()
 
-    int crossings = result[0];
-    return crossings;
+    if (crossings[0] > 0)
+      geom.geometryChanged();//applies to call component Geometries
+    return crossings[0];
   }
 
   /** See {@link #unwrapDateline(Geometry)}. */
@@ -257,6 +363,7 @@ public class JtsGeometry implements Shape {
     LineString exteriorRing = poly.getExteriorRing();
     int cross = unwrapDateline(exteriorRing);
     if (cross > 0) {
+      //TODO TEST THIS! Maybe bug if doesn't cross but is in another page?
       for(int i = 0; i < poly.getNumInteriorRing(); i++) {
         LineString innerLineString = poly.getInteriorRingN(i);
         unwrapDateline(innerLineString);
@@ -267,7 +374,6 @@ public class JtsGeometry implements Shape {
           shiftGeomByX(innerLineString, 360);
         }
       }
-      poly.geometryChanged();
     }
     return cross;
   }
@@ -300,6 +406,7 @@ public class JtsGeometry implements Shape {
       }
       if (shiftXPage != 0)
         cseq.setOrdinate(i, CoordinateSequence.X, thisX);
+      prevX = thisX;
     }
     if (lineString instanceof LinearRing) {
       assert cseq.getCoordinate(0).equals(cseq.getCoordinate(size-1));
@@ -309,8 +416,6 @@ public class JtsGeometry implements Shape {
     //Unfortunately we are shifting again; it'd be nice to be smarter and shift once
     shiftGeomByX(lineString, shiftXPageMin * -360);
     int crossings = shiftXPageMax - shiftXPageMin;
-    if (crossings > 0)
-      lineString.geometryChanged();
     return crossings;
   }
 
@@ -348,7 +453,8 @@ public class JtsGeometry implements Shape {
       return geom;
     assert geom.isValid() : "geom";
 
-    //TODO support geom's that start at negative pages; will avoid need to previously shift in unwrapDateline(geom).
+    //TODO opt: support geom's that start at negative pages --
+    // ... will avoid need to previously shift in unwrapDateline(geom).
     List<Geometry> geomList = new ArrayList<Geometry>();
     //page 0 is the standard -180 to 180 range
     for (int page = 0; true; page++) {
